
#include "stm32f10x.h"
#include "stdlib.h"
#include "DDH47.h"

// "OPTS" : damit beim Debug-Kompilat erstens "fMHz" reicht und zweitens Programm noch im Ram (statt Flash) passt
#define OPTS __attribute__((optimize("-Os")))

#define my_assert(Condition, InfoVal)  if(!(Condition) && _State->Error==_ErrOk) _State->Error=InfoVal;


static TDDH47State*   _State   = 0; // separat zu _Control, damit _Control "verborgener" bleibt
static TDDH47Control* _Control = 0;

//################################################### TimedAdc_Init
static int32_t* OPTS TimedAdc_Init(TTimedAdc* s, int32_t ppm)
{
  // - Quarzabweichung ausgleichbar
  // - Grobabgleich per ppm-Parameter in 1 ppm Schrtten
  //   (positiver Wert: Quarz schwingt schneller als Nominalfrequenz)
  // - Feinabgleich per TimedAdc_Tweak() in 1/256 ppm Schritten
  // - verursacht ca. +-2 Trgerphasenrauschen bei fCPU=16MHz und fCarrier=77.5kHz (=77.5kHz/16MHz*360)
  // - der DMA darf weiterlaufen
  int32_t fADC = fCarrier*ADCSamplesPerPeriod;
  int n0 = fMHz*(1000*1000+ppm)/fADC; // STM32F030 DIV?
  int a0 = fMHz*(1000*1000+ppm)-fADC*n0;
  my_assert(n0<=254, _ErrTimedAdc);
  int a1 = fADC;
  int w = -a1;
  int ns = 0;
  for(int i=0; i<AdcBufCnt; i++)
  {
    int n = n0;
    w += a0;
    if(w>0)
    {
      n = n0+1;
      w -= a1;
    }
    s->TimArr[i] = n-1;
    ns += n;
  }
  s->n  = n0;
  s->a0 = fMHz*(1000*1000+ppm)*AdcBufCnt-fADC*ns;
  s->a1 = fADC;
  s->w  = -s->a1;
  return &s->d_ppm_L10; // (darf von auen manipuliert werden)
}

//################################################### TimedAdc_Tweak
static void OPTS TimedAdc_Tweak(TTimedAdc* s)
{
  // je TimedAdc.TimArr[]-DMA-Durchlauf _genau_einmal_ aufrufen:
  // Der Feinabgleichsparameter "d_ppm_L10" darf jederzeit von auen
  // gendert.
  int n = s->n; // Standard (kein extra CPU-Takt)
  s->w += s->a0;
  if(s->w>0)
  {
    n = s->n+1; // Sonderfall: ein CPU-Takt extra
    s->w -= s->a1;
  }
  s->TimArr[0] = n-1;
  // neuen Korrekturwert anwenden?
  if(s->d_ppm_L10!=0)
  {
    int a0 = s->a0+(s->d_ppm_L10*AdcBufCnt*fMHz)/1024;
    s->d_ppm_L10 = 0;
    int d = 0;
    if(a0<0) d=+1;
    if(a0>=s->a1) d=-1;
    s->n -= d;
    s->a0 = a0+d*s->a1;
  }
}

//################################################### DoDCT
static void OPTS DoDCT(int16_t* Arr, int32_t PeriodCnt, int32_t* ValC, int32_t* ValS)
{
  // berlagerung mit Amplitude von 256
  my_assert(ADCSamplesPerPeriod>=3 && ADCSamplesPerPeriod<=12, _ErrDCT);
  int32_t SumC = *ValC;
  int32_t SumS = *ValS;
  do
  {
    int32_t v;
    // (Code wurde automatisiert erstellt)
    if(ADCSamplesPerPeriod==3)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 2
      SumC -= v*128;
      SumS -= v*221;
    }
    if(ADCSamplesPerPeriod==4)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*  0;
      SumS += v*256;
      v = *Arr++;    // 2
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 3
      SumC += v*  0;
      SumS -= v*256;
    }
    if(ADCSamplesPerPeriod==5)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v* 79;
      SumS += v*243;
      v = *Arr++;    // 2
      SumC -= v*207;
      SumS += v*150;
      v = *Arr++;    // 3
      SumC -= v*207;
      SumS -= v*150;
      v = *Arr++;    // 4
      SumC += v* 79;
      SumS -= v*243;
    }
    if(ADCSamplesPerPeriod==6)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*128;
      SumS += v*221;
      v = *Arr++;    // 2
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 3
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 4
      SumC -= v*128;
      SumS -= v*221;
      v = *Arr++;    // 5
      SumC += v*128;
      SumS -= v*221;
    }
    if(ADCSamplesPerPeriod==7)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*159;
      SumS += v*200;
      v = *Arr++;    // 2
      SumC -= v* 57;
      SumS += v*249;
      v = *Arr++;    // 3
      SumC -= v*230;
      SumS += v*111;
      v = *Arr++;    // 4
      SumC -= v*230;
      SumS -= v*111;
      v = *Arr++;    // 5
      SumC -= v* 57;
      SumS -= v*249;
      v = *Arr++;    // 6
      SumC += v*159;
      SumS -= v*200;
    }
    if(ADCSamplesPerPeriod==8)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*181;
      SumS += v*181;
      v = *Arr++;    // 2
      SumC += v*  0;
      SumS += v*256;
      v = *Arr++;    // 3
      SumC -= v*181;
      SumS += v*181;
      v = *Arr++;    // 4
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 5
      SumC -= v*181;
      SumS -= v*181;
      v = *Arr++;    // 6
      SumC += v*  0;
      SumS -= v*256;
      v = *Arr++;    // 7
      SumC += v*181;
      SumS -= v*181;
    }
    if(ADCSamplesPerPeriod==9)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*196;
      SumS += v*164;
      v = *Arr++;    // 2
      SumC += v* 44;
      SumS += v*252;
      v = *Arr++;    // 3
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 4
      SumC -= v*240;
      SumS += v* 87;
      v = *Arr++;    // 5
      SumC -= v*240;
      SumS -= v* 87;
      v = *Arr++;    // 6
      SumC -= v*128;
      SumS -= v*221;
      v = *Arr++;    // 7
      SumC += v* 44;
      SumS -= v*252;
      v = *Arr++;    // 8
      SumC += v*196;
      SumS -= v*164;
    }
    if(ADCSamplesPerPeriod==10)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*207;
      SumS += v*150;
      v = *Arr++;    // 2
      SumC += v* 79;
      SumS += v*243;
      v = *Arr++;    // 3
      SumC -= v* 79;
      SumS += v*243;
      v = *Arr++;    // 4
      SumC -= v*207;
      SumS += v*150;
      v = *Arr++;    // 5
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 6
      SumC -= v*207;
      SumS -= v*150;
      v = *Arr++;    // 7
      SumC -= v* 79;
      SumS -= v*243;
      v = *Arr++;    // 8
      SumC += v* 79;
      SumS -= v*243;
      v = *Arr++;    // 9
      SumC += v*207;
      SumS -= v*150;
    }
    if(ADCSamplesPerPeriod==11)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*215;
      SumS += v*138;
      v = *Arr++;    // 2
      SumC += v*106;
      SumS += v*232;
      v = *Arr++;    // 3
      SumC -= v* 37;
      SumS += v*253;
      v = *Arr++;    // 4
      SumC -= v*167;
      SumS += v*193;
      v = *Arr++;    // 5
      SumC -= v*245;
      SumS += v* 72;
      v = *Arr++;    // 6
      SumC -= v*245;
      SumS -= v* 72;
      v = *Arr++;    // 7
      SumC -= v*167;
      SumS -= v*193;
      v = *Arr++;    // 8
      SumC -= v* 37;
      SumS -= v*253;
      v = *Arr++;    // 9
      SumC += v*106;
      SumS -= v*232;
      v = *Arr++;    // 10
      SumC += v*215;
      SumS -= v*138;
    }
    if(ADCSamplesPerPeriod==12)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*221;
      SumS += v*127;
      v = *Arr++;    // 2
      SumC += v*128;
      SumS += v*221;
      v = *Arr++;    // 3
      SumC += v*  0;
      SumS += v*256;
      v = *Arr++;    // 4
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 5
      SumC -= v*221;
      SumS += v*127;
      v = *Arr++;    // 6
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 7
      SumC -= v*221;
      SumS -= v*127;
      v = *Arr++;    // 8
      SumC -= v*128;
      SumS -= v*221;
      v = *Arr++;    // 9
      SumC += v*  0;
      SumS -= v*256;
      v = *Arr++;    // 10
      SumC += v*128;
      SumS -= v*221;
      v = *Arr++;    // 11
      SumC += v*221;
      SumS -= v*127;
    }
    if(ADCSamplesPerPeriod==8 && 1==2)
    {
      enum
      {
        ScaleSin00 =   0,
        ScaleSin45 = 256,
        ScaleSin90 = 362  // =ScaleSin45/sin(45)
      };
      int v;
      //                    // #   cos   sin
      v = *Arr++;
      SumC += v*ScaleSin90; // 0  +100     0
      SumS += v*ScaleSin00;
      v = *Arr++;
      SumC += v*ScaleSin45; // 1   +71   +71
      SumS += v*ScaleSin45;
      v = *Arr++;
      SumC += v*ScaleSin00; // 2     0  +100
      SumS += v*ScaleSin90;
      v = *Arr++;
      SumC -= v*ScaleSin45; // 3   -71   +71
      SumS += v*ScaleSin45;
      v = *Arr++;
      SumC -= v*ScaleSin90; // 4  -100     0
      SumS += v*ScaleSin00;
      v = *Arr++;
      SumC -= v*ScaleSin45; // 5   -71   -71
      SumS -= v*ScaleSin45;
      v = *Arr++;
      SumC += v*ScaleSin00; // 6     0  -100
      SumS -= v*ScaleSin90;
      v = *Arr++;
      SumC += v*ScaleSin45; // 7   +71   -71
      SumS -= v*ScaleSin45;
    }
  } while(--PeriodCnt>0);
  *ValC = SumC;
  *ValS = SumS;
}

//############################################################## LogBase2
static inline int32_t __attribute__((always_inline)) LogBase2(uint32_t v)
{
  // "Find the log base 2 of an N-bit integer in O(lg(N)) operations with multiply and lookup"
  // ("https://graphics.stanford.edu/~seander/bithacks.html")
  int r;      // result goes here
  static const int8_t MultiplyDeBruijnBitPosition[32] =
  {
    0, 9, 1, 10, 13, 21, 2, 29, 11, 14, 16, 18, 22, 25, 3, 30,
    8, 12, 20, 28, 15, 17, 24, 7, 19, 27, 23, 6, 26, 5, 4, 31
  };
  v |= v >> 1; // first round down to one less than a power of 2
  v |= v >> 2;
  v |= v >> 4;
  v |= v >> 8;
  v |= v >> 16;
  r = MultiplyDeBruijnBitPosition[(uint32_t)(v * 0x07C4ACDDU) >> 27];
  return r;
}

//############################################################## uint8_atan2
static uint8_t OPTS uint8_atan2(int32_t y, int32_t x)
{
  // berechnet int(Period*atan2(y, x)/2/4/atan(1))&(Period-1)
  // -  +-1 Abweichnung bei Period==256 und AtnCnt==32
  enum
  {
    Period = 256,
    AtnShift = 5,
    AtnCnt = 1<<AtnShift
  };
  static const uint8_t AtnTab[AtnCnt] =
  {
    // AtnTab[i] = int(Period*atan(1.0*i/(AtnCnt-1))/2/4/atan(1)+.5);
    0x00, 0x01, 0x03, 0x04, 0x05, 0x07, 0x08, 0x09,
    0x0a, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12,
    0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a,
    0x1b, 0x1c, 0x1c, 0x1d, 0x1e, 0x1f, 0x1f, 0x20,
  };
  int xAbs = x<0 ? -x : +x;
  int yAbs = y<0 ? -y : +y;
  if(xAbs>0x7fffffff>>AtnShift || xAbs>0x7fffffff>>AtnShift)
  {
    xAbs >>= AtnShift;
    yAbs >>= AtnShift;
  }
  int phi = AtnTab[xAbs<yAbs ? AtnCnt*xAbs/(yAbs+1) : AtnCnt*yAbs/(xAbs+1)]; // STM32F030 DIV?
  if(xAbs<yAbs) phi=Period/4-phi;
  if(x<0) phi=Period/2-phi;
  if(y<0) phi=Period-phi;
  return phi;
}

//############################################################## int32_sqrt
static int32_t OPTS int32_sqrt(int32_t x)
{
  // https://en.wikipedia.org/wiki/Integer_square_root
  int shift = 2;
  while((x>>shift)!=0) shift+=2;
  int y = 0;
  for(; shift>=0; shift-=2)
  {
    y <<= 1;
    int32_t v = y+1;
    if(v*v<=(x>>shift)) y=v;
  }
  return y;
}

enum
{
  SinTabAmp = 128,
  SinTabCnt = 256,
  CarrierAmpMax = 0x7fffffff/SinTabAmp,
};

static const int8_t SinTab[SinTabCnt] =
{
  // SinTab[i] = int(SinTabAmp*sin(i*2*4*atan(1)/SinTabCnt));
    0,    3,    6,    9,   12,   15,   18,   21,
   24,   27,   30,   33,   36,   39,   42,   45,
   48,   51,   54,   57,   59,   62,   65,   67,
   70,   73,   75,   78,   80,   82,   85,   87,
   89,   91,   94,   96,   98,  100,  102,  103,
  105,  107,  108,  110,  112,  113,  114,  116,
  117,  118,  119,  120,  121,  122,  123,  123,
  124,  125,  125,  126,  126,  126,  126,  126,
  127,  126,  126,  126,  126,  126,  125,  125,
  124,  123,  123,  122,  121,  120,  119,  118,
  117,  116,  114,  113,  112,  110,  108,  107,
  105,  103,  102,  100,   98,   96,   94,   91,
   89,   87,   85,   82,   80,   78,   75,   73,
   70,   67,   65,   62,   59,   57,   54,   51,
   48,   45,   42,   39,   36,   33,   30,   27,
   24,   21,   18,   15,   12,    9,    6,    3,
    0,   -3,   -6,   -9,  -12,  -15,  -18,  -21,
  -24,  -27,  -30,  -33,  -36,  -39,  -42,  -45,
  -48,  -51,  -54,  -57,  -59,  -62,  -65,  -67,
  -70,  -73,  -75,  -78,  -80,  -82,  -85,  -87,
  -89,  -91,  -94,  -96,  -98, -100, -102, -103,
  -105, -107, -108, -110, -112, -113, -114, -116,
  -117, -118, -119, -120, -121, -122, -123, -123,
  -124, -125, -125, -126, -126, -126, -126, -126,
  -127, -126, -126, -126, -126, -126, -125, -125,
  -124, -123, -123, -122, -121, -120, -119, -118,
  -117, -116, -114, -113, -112, -110, -108, -107,
  -105, -103, -102, -100,  -98,  -96,  -94,  -91,
  -89,  -87,  -85,  -82,  -80,  -78,  -75,  -73,
  -70,  -67,  -65,  -62,  -59,  -57,  -54,  -51,
  -48,  -45,  -42,  -39,  -36,  -33,  -30,  -27,
  -24,  -21,  -18,  -15,  -12,   -9,   -6,   -3,
};

//############################################################## BMDecode
static char OPTS BMDecode(bool* SD_FigureShift, uint8_t BMCode)
{
  char res = BMCodeTable[2*BMCode + (*SD_FigureShift ? 1 : 0)];
  if(BMCode==27) *SD_FigureShift=true;
  if(BMCode==31) *SD_FigureShift=false;
  return res;
}

//############################################################## HBI
static inline int32_t __attribute__((always_inline)) HBI(int32_t i)
{
  if(i<0) i+=DDH47_HBCnt;
  if(i>=DDH47_HBCnt) i-=DDH47_HBCnt;
  return i;
}

//############################################################## IRQ_ProcessSample
void OPTS IRQ_ProcessSample(TDDH47Control* s, int32_t CurC, int32_t CurS)
{
  // FSKAmp: Amplitudendifferenz der FSK-Frequenzen ber eine Bitlnge gemittelt
  int32_t FSKAmp = 0;
  int32_t FSK_C[2];
  int32_t FSK_S[2];
  {
    int32_t dC = CurC-s->PrevC;
    int32_t dS = CurS-s->PrevS;
    s->PrevC = CurC;
    s->PrevS = CurS;
    // Platz fr Faktor SinTabAmp schaffen
    const int DownShift = LogBase2(2*Sample_MaxValue/CarrierAmpMax-1);
    dC >>= DownShift;
    dS >>= DownShift;
    // s->dPhiFSK: aktuelle Phasenverschiebung der FSK-Schwingung gegenber der
    // Phase der Mittenfrequenz; geshiftet um 24; Perodenlnge=256
    s->dPhiFSK += (uint32_t)(256.0*256*65536*PeriodsPerSample*dfCarrierMinMax/2/fCarrier);
    int CosVal = SinTab[(1*0x40000000+s->dPhiFSK)>>24];
    int SinVal = SinTab[(0*0x40000000+s->dPhiFSK)>>24];
    int AvrPrev = s->AvrCur;
    s->AvrCur++;
    if(s->AvrCur==DDH47_AvrCnt) s->AvrCur=0;
    for(int BitVal=0; BitVal<2; BitVal++)
    {
      // C, S fr Mittelwert
      SinVal = -SinVal; // -> "- s->dPhiFSK"
      int32_t* Arr = BitVal==0 ? s->AvrNeg : s->AvrPos;
      my_assert(sqrt(2)*SMPPerSecond/Baud<SinTabAmp, 91); // proforma
      int32_t SumC = Arr[2*AvrPrev+0] + (+dC*CosVal + dS*SinVal)/SinTabAmp;
      int32_t SumS = Arr[2*AvrPrev+1] + (-dC*SinVal + dS*CosVal)/SinTabAmp;
      FSK_C[BitVal] = SumC-Arr[2*s->AvrCur+0];
      FSK_S[BitVal] = SumS-Arr[2*s->AvrCur+1];
      int C = abs(FSK_C[BitVal]);
      int S = abs(FSK_S[BitVal]);
      Arr[2*s->AvrCur+0] = SumC;
      Arr[2*s->AvrCur+1] = SumS;
      // Amp
      int nShift = 0;
      while(C>=(int32_t)(0x7fff/sqrt(2)) || S>=(int32_t)(0x7fff/sqrt(2)))
      {
        C >>= 1;
        S >>= 1;
        nShift++;
      }
      FSKAmp = (int32_sqrt(C*C+S*S)<<nShift)-FSKAmp;
    }
  }
  s->FSKAmpSum += FSKAmp;
  s->FSKAmpCnt++;

  // Dekoder #1 : Ein einfacher Dekoder.
  // - 1 Startbit
  // - 5 Datenbits
  // - 1.5 Stoppbits
  {
    s->SD_Balance += PeriodsPerSample*Baud*4;
    if(s->SD_iQuarterBit>=27 && FSKAmp>0)
    {
      s->SD_Balance = 0;
    }
    if(s->SD_Balance>=fCarrier)
    {
      if(s->SD_iQuarterBit>=27) s->SD_iQuarterBit=0;
      s->SD_Balance -= fCarrier;
      if((s->SD_iQuarterBit&3)==1)
      {
        s->SD_Byte >>= 1;
        if(FSKAmp>0) s->SD_Byte|=1<<6;
        if(s->SD_iQuarterBit==6*4+1)
        {
          _State->TriteDecode = BMDecode(&s->SD_FigureShift, (s->SD_Byte>>1)&0x1f);
        }
      }
      s->SD_iQuarterBit++;
    }
  }

  // Dekoder #2 : Ein aufwndiger und nicht ganz korrekter Dekoder.
  // - 1 Startbit
  // - 5 Datenbits
  // - 1.5 Stoppbits
  // - weitere, aber (vermutlich) nicht spezifizierte, Vorraussetzungen
  //   - alle Bits beginnen auf einem 0.5 Bitraster
  //   - eine Zeichenlcke hat maximal 7 Bits Lnge
  //   - es werden immer mindestens "DDH47_SyncCharCnt" Zeichen lckenlos
  //     gesendet

  // Amplitude zum Histogramm zur Bestimmung der Halbbitphase hinzufgen
  {
    int32_t dAmp = abs(FSKAmp-s->AmpVal[s->AmpCur]);
    s->AmpVal[s->AmpCur] = FSKAmp;
    s->AmpCur++;
    if(s->AmpCur==DDH47_AmpCnt) s->AmpCur=0;
    // - Es wird davon ausgegangen (was unzulssig ist), dass alle Bitanfnge
    //   auf einem dt=0.5/Baud Raster liegen
    // - bzw. beim Bitanfang (s->ScaledBitCur%(DDH47_BitScale/2)) konstant ist
    // - die Konstante wird per Histogramm "ScaledBitHist[]" bestimmt
    // - ein Histogramm ber 0.5 Bit ist zu sehr verwischt -> deswegen 7.5 Bit
    const int d = (int)(1.0*DDH47_BitPhaseTau_s*SMPPerSecond/(DDH47_BitScale*DDH47_HalbBitsPerChar/2));
    s->ScaledBitHist[s->ScaledBitCur] += (dAmp-s->ScaledBitHist[s->ScaledBitCur])/d;
  }

  bool NewScaledBit = false;

  // aktuelle Bitposition ScaledBitCur aktualisieren
  s->ScaledBitBalance += DDH47_BitScale*PeriodsPerSample*Baud;
  if(s->ScaledBitBalance>=fCarrier)
  {
    NewScaledBit = true;
    s->ScaledBitBalance -= fCarrier;
    s->ScaledBitCur++;
    my_assert((DDH47_BitScale&1)==0, 92);
    if(s->ScaledBitCur==DDH47_ScaledBitCnt) s->ScaledBitCur=0;
    int Phase = (DDH47_BitScale+s->ScaledBitCur-s->ScaledBitPhase)%(DDH47_BitScale/2);
    if(Phase==(DDH47_BitScale/4)) s->ScaledBitPhase=s->ScaledBitPhaseNext;
    // Phase fr neues HBBit
    if(Phase==0)
    {
      s->HBArr[s->HBCur].Phase[0] = uint8_atan2(FSK_S[0], FSK_C[0]);
      s->HBArr[s->HBCur].Phase[1] = uint8_atan2(FSK_S[1], FSK_C[1]);
    }
    // Messung fr neues HBBit fertig
    if(Phase==DDH47_FSKAmpAvr_dn)
    {
      s->HBArr[s->HBCur].Amp = DDH47_FSKAmpAvr_dn==0 ? FSKAmp : s->FSKAmpSum/s->FSKAmpCnt;
      s->HBCur++;
      if(s->HBCur==DDH47_HBCnt) s->HBCur=0;
    }
    // Init fr neues HBBit
    if(Phase>=DDH47_FSKAmpAvr_dn && Phase<=DDH47_BitScale/2-DDH47_FSKAmpAvr_dn)
    {
      s->FSKAmpSum = 0;
      s->FSKAmpCnt = 0;
    }
    if(Phase==0) s->State=1;
    _State->HalfBitCnt++;
  }

  // in ScaledBitHist[] die Position vom Maximum suchen
  if(NewScaledBit)
  {
    // Maximum der Summe von "DDH47_PhaseSumCnt"-Array-Werten
    int Pos = DDH47_ScaledBitCnt+s->ScaledBitCur-DDH47_PhaseSumCnt-1;
    my_assert(DDH47_PhaseSumCnt<DDH47_BitScale, 93); // sonst Summenberlauf
    s->ScaledBitPhaseSum -= s->ScaledBitHist[(Pos+0*DDH47_PhaseSumCnt)%DDH47_ScaledBitCnt]/DDH47_BitScale;
    s->ScaledBitPhaseSum += s->ScaledBitHist[(Pos+1*DDH47_PhaseSumCnt)%DDH47_ScaledBitCnt]/DDH47_BitScale;
    if(s->ScaledBitPhaseSum>s->ScaledBitPhaseMax)
    {
      s->ScaledBitPhasePos = Pos;
      s->ScaledBitPhaseMax = s->ScaledBitPhaseSum;
    }
    if(s->ScaledBitCur==0)
    {
      int p = s->ScaledBitPhasePos+1-DDH47_PhaseSumCnt/2;
      s->ScaledBitPhaseNext = (DDH47_BitScale/2+p)%(DDH47_BitScale/2);
      s->ScaledBitPhaseMax = 0;
    }
  }

  // Startbitgtewert "SyncArr[]" frs Halbbit berechnen
  if(s->State==DDH47_State1_CalcChrQuality)
  {
    int HBOfs = DDH47_HBCnt+s->HBCur-DDH47_HalbBitsPerChar-1;
    int qSum = 0;
    for(int p=0; p<DDH47_HalbBitsPerChar-1; p++)
    {
      int a0 = s->HBArr[HBI(HBOfs+p+0)].Amp;
      int a2 = s->HBArr[HBI(HBOfs+p+2)].Amp;
      int q = (abs(a0-a2)-abs(a0+a2));
      if(q<0) q=0;
      if((p&1)!=0) q=-q;
      qSum += q;
    }
    if(qSum<0) qSum=0;
    if(s->HBArr[(HBOfs+0)%DDH47_HBCnt].Amp<0) qSum=0; // letztes 2/3 vom 1.5er Stopbit
    if(s->HBArr[(HBOfs+2)%DDH47_HBCnt].Amp>0) qSum=0; // Startbit
    s->HBArr[(HBOfs+0)%DDH47_HBCnt].StartAmp = qSum;
  }

  // aus Startbitgtewerten mehrerer Zeichen beste Starthalbbitphase ermitteln
  if(s->State==DDH47_State2_NextCharPhaseA)
  {
    int Sum = 0;
    int Pos = (DDH47_HBCnt+s->HBCur-DDH47_HalbBitsPerChar-1)%DDH47_HalbBitsPerChar;
    if(Pos==0)
    {
      s->CharPhaseNext = s->SyncMaxPos;
      s->SyncMaxVal = 0;
    }
    for(int i=0; i<DDH47_SyncCharCnt; i++)
    {
      Sum += s->HBArr[Pos+i*DDH47_HalbBitsPerChar].StartAmp;
    }
    if(Sum>s->SyncMaxVal)
    {
      s->SyncMaxVal = Sum;
      s->SyncMaxPos = Pos;
    }
  }

  // Zeichen dekodieren
  if(s->State==DDH47_State3_NextCharPhaseB)
  {
    // gegebenenfalls aktuelle Starthalbbitphase "CharPhase" (0..14) erneuern
    bool DoDecode = (s->HBCur%DDH47_HalbBitsPerChar)==0;
    if(DoDecode && s->CharPhase!=s->CharPhaseNext)
    {
      int nOk = 0;
      for(int iChar=0; nOk>=0 && iChar<DDH47_SyncCharCnt-1; iChar++)
      {
        int pp = DDH47_HBCnt+s->HBCur+iChar*DDH47_HalbBitsPerChar;
        int p0 = pp+s->CharPhase;
        int p1 = pp+s->CharPhaseNext;
        if(p1<p0) p1+=DDH47_HalbBitsPerChar;
        int v0 = s->HBArr[p0%DDH47_HBCnt].StartAmp;
        int v1 = s->HBArr[p1%DDH47_HBCnt].StartAmp;
        nOk += 1*v1>5*v0 ? +1 : -1;
      }
      if(nOk>0)
      {
        DoDecode = s->CharPhaseNext>s->CharPhase;
        s->CharPhase = s->CharPhaseNext;
      }
    }
    // Zeichen dekodieren
    if(DoDecode)
    {
      int p0 = (s->HBCur+s->CharPhase+4)%DDH47_HBCnt;
      int Byte = 0;
      for(int i=0; i<5; i++)
      {
        // if(((s->HBArr[p0].StartAmp)&1)!=0) Byte|=1<<i;
        if((s->HBArr[p0].Amp)>0) Byte|=1<<i;
        p0 += 2;
        if(p0>=DDH47_HBCnt) p0-=DDH47_HBCnt;
      }
      _State->FancyDecode = BMDecode(&s->BMCodeFigureShift, Byte);
    }
  }

  // Infos ber Signal, Rauschen und STM32-Quarzabweichnung berechnen
  if(s->State==DDH47_State4_Info)
  {
    // Hierzu verwendbar sind nur Amplitudenwerte bei denen sich die
    // Senderfrequenz nicht gendert hat. Falls mehrere Halbits gleich sind,
    // dann trifft das fr die mittleren wahrscheinlich zu.
    int HBits = 0;
    for(int i=0; i<4; i++)
    {
      if(s->HBArr[(DDH47_HBCnt+s->HBCur-1-i)%DDH47_HBCnt].Amp>0) HBits|=1<<i;
    }
    bool IsSingleHB = (HBits& 7)==0 || (HBits& 7)== 7;
    bool IsDoubleHB = (HBits&15)==0 || (HBits&15)==15;
    if(IsSingleHB)
    {
      s->InfoCurAmp = s->HBArr[(DDH47_HBCnt+s->HBCur-1-1)%DDH47_HBCnt].Amp;
    }
    int a = s->InfoCurAmp;
    int BitVal = a>0 ? 1 : 0;
    int d = abs(a-s->InfoPrevAmp[BitVal]);
    if(IsSingleHB) s->InfoPrevAmp[BitVal]=a;
    a = abs(a);
    // _State->SN_uV
    {
      s->InfoSignal += (a-s->InfoSignal)/(2*Baud*DDH47_InfoSignalTau_s);
      const int DownShift = LogBase2(2*Sample_MaxValue/CarrierAmpMax-1);
      const int k = (int32_t)((1.0*DDH47_AvrCnt*Sample_MaxValue/(1<<DownShift)) / (1e6*(4/M_PI)*(3.3/2/sqrt(2))));
      _State->SN_uV = s->InfoSignal/k;
    }
    // _State->SNR_Promille
    {
      if(d>a) d=a;
      while(a>=0x7fff)
      {
        d >>= 1;
        a >>= 1;
      }
      int d2 = d*d;
      int a2 = a*a;
      // Verhltnis a zu d um 1e6 (Promille^2) vergrern
      {
        int n = 20; // 2^20 "=" 1e6
        for(; n>0 && d2<0x40000000; n--)
        {
          d2 <<= 1;
        }
        for(; n>0; n--)
        {
          a2 >>= 1;
        }
      }
      int v2 = a2>0 ? d2/a2 : 0;
      s->InfoSNR2 += (v2-s->InfoSNR2)/(2*Baud*DDH47_InfoNoiseTau_s);
      int NSR_E3 = int32_sqrt(s->InfoSNR2);  // = 1000*Noise/Signal
      if(NSR_E3<10) NSR_E3=10;
      _State->SNR = 1000/NSR_E3;
    }
    // _State->dfHSE_ppm
    if(IsDoubleHB)
    {
      // dPhi: Phasennderung ber ein Halbbit; Sollwert: 0

      int8_t Phi0 = s->HBArr[(DDH47_HBCnt+s->HBCur-1-2)%DDH47_HBCnt].Phase[BitVal];
      int8_t Phi1 = s->HBArr[(DDH47_HBCnt+s->HBCur-1-1)%DDH47_HBCnt].Phase[BitVal];
      int8_t dPhi = Phi1-Phi0;
      s->Info_dfHSE += (dPhi*1024-s->Info_dfHSE)/10;
      const int c = (int32_t)(1024*(1e-6*fCarrier*256/(2*Baud)));
      _State->dfHSE_ppm = s->Info_Cur_ppm+s->Info_dfHSE/c;
    }
  }

  s->State++;
}

//################################################### TIM3_IRQHandler
void OPTS TIM3_IRQHandler()
{
  TDDH47Control* s = _Control;

  // Init Zeitpunkt fr nchsten IRQ
  {
    // auerdem wird der neue IRQ-Zeitpunkt um die Dauer einer DDH-Periode verrauscht:
    int PrevNoiseVal = ((s->Rand>>16)*ADCSamplesPerPeriod)>>16; // Zufallszahl 0..(ADCSamplesPerPeriod-1)
    s->Rand = s->Rand*214013+2531011;
    int NewNoiseVal = ((s->Rand>>16)*ADCSamplesPerPeriod)>>16; // Zufallszahl 0..(ADCSamplesPerPeriod-1)
    TIM3->CCR1 = TIM3->CCR1+AdcCntPerSample/IRQPerSample+NewNoiseVal-PrevNoiseVal;
    // durch das _bedingte_ Lschen wird auch bei 30ms blokiertem IRQ eine Eskalation verhindert:
    if((int16_t)(TIM3->CCR1-TIM3->CNT)>1) TIM3->SR=~TIM_SR_CC1IF;
  }

  // DCT von den neuen ADC-Samples durchfhren
  {
    DoDCT(s->TimedAdc.ValArr+s->AdcBufOfs, AdcCntPerSample/IRQPerSample/ADCSamplesPerPeriod, &s->CurC, &s->CurS);
    my_assert(PeriodsPerSample%IRQPerSample==0, _ErrAdcBufSize);
    // wenn folgendes my_assert() nur _vereinzelt_ fehlschlgt, dann drfte es entfernt werden
    my_assert((uint32_t)(AdcBufCnt-DMA1_Channel2->CNDTR-s->AdcBufOfs)>=AdcCntPerSample/IRQPerSample, _ErrAdcBufOverrun);
    s->AdcBufOfs += AdcCntPerSample/IRQPerSample;
    if(s->AdcBufOfs==AdcBufCnt) s->AdcBufOfs=0;
    s->IRQCur++;
  }

  // Sample verarbeiten
  if(s->IRQCur==IRQPerSample)
  {
    s->IRQCur = 0;
    IRQ_ProcessSample(s, s->CurC, s->CurS);
    if(DbgSample_Cnt!=0)
    {
      static int32_t DbgSample_Arr[DbgSample_Cnt];
      static int32_t DbgSample_Ofs = 0;
      DbgSample_Arr[(DbgSample_Ofs++)&(DbgSample_Cnt-1)] = s->CurC;
      DbgSample_Arr[(DbgSample_Ofs++)&(DbgSample_Cnt-1)] = s->CurS;
      TIM2->CCR3 = (uint32_t)DbgSample_Arr; // fr VisualStudio sichtbar machen (Missbrauch)
      TIM2->CCR4 = (uint32_t)DbgSample_Ofs; // fr VisualStudio sichtbar machen (Missbrauch)
    }
  }

  // Feinabgleich von fADC
  if((DMA1->ISR&DMA_ISR_TCIF2)!=0 && DMA1_Channel2->CNDTR<AdcBufCnt-1)
  {
    DMA1->IFCR = DMA_IFCR_CTCIF2;
    TimedAdc_Tweak(&s->TimedAdc);
  }
}

//################################################### DDH47_Init
void OPTS DDH47_Init(TDDH47Control* Control, TDDH47State* State, int32_t ppm)
{
  _Control = Control;
  _State   = State;

  for(int i=0; i<sizeof(*_Control); i++)
  {
    ((uint8_t*)_Control)[i] = 0;
  }
  for(int t=0; t<sizeof(*_State); t++)
  {
    ((uint8_t*)_State)[t] = 0;
  }

  // PA7: Analog
  RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
  GPIOA->CRL = (GPIOA->CRL&0x0fffffff)|0x00000000;

  // ADC_Frequenz einstellen (TimedAdc.TimArr[] definieren)
  TimedAdc_Init(&Control->TimedAdc, ppm);
  Control->Info_Cur_ppm = ppm;

  // TIM2
  // - fUpdate=fADC
  // - Trigger fr ADC-Sample-Start
  // - Trigger fr TIM3-Counter
  RCC->APB1ENR |= RCC_APB1ENR_TIM2EN;
  TIM2->ARR = Control->TimedAdc.TimArr[0]; // Achtung: wird per DMA1_Channel2 manipuliert!
  TIM2->CCR1 = 10;                         // TIM3-Zhler-Trigger
  TIM2->CCR2 = 10;                         // ADC-Start-Trigger
  TIM2->CCER = TIM_CCER_CC2E;              // CC2 output enabled (fr ADC-Start-Trigger)
  TIM2->CCMR1 = TIM_CCMR1_OC2M_0*7;        // 7: PWM mode 2 - active TIMx_CNT>=TIMx_CCR1 (fr ADC-Start-Trigger)
  TIM2->CR2 = TIM_CR2_MMS_0*3;             // 3: Master Trigger Output "Compare CC1IF" (fr TIM3-Trigger)
  TIM2->DIER = TIM_DIER_UDE;               // Update Event (fr DMA1_Channel2-Trigger)

  // TIM3
  // - zhlt ADC-Samples
  // - hilft Strungen zu reduzieren (hilft Rauschen zu erzeugen)
  //   Beispiel:
  //   Bei 64MHz-STM32-Frequenz und STM32-Sleeps ("WFI()") gibts es
  //   deutliche Stromnderungen zwischen Sleep und der Bearbeitung in IRQ_ProcessSample().
  //   Da die Stromnderung immer zur gleichen Phase des Mischsignals
  //   (DoDCT8()) einsetzt, tritt als Folge bei Carrier() auch bei
  //   fADC/ADCSamplesPerPeriod ein "Phantom-Sender" auf. Ein schwaches
  //   Trgersignal kann dann nicht sicher erkannt werden. Mit Hilfe von TIM3
  //   kann der Zeitpunkt fr den IRQ und damit die Phase der Strung verndert
  //   werden.
  RCC->APB1ENR |= RCC_APB1ENR_TIM3EN;
  TIM3->SMCR |= TIM_SMCR_TS_0*1;      // 1: Slave Trigger "Internal Trigger 1 (ITR1)" -> TIM2
  TIM3->SMCR |= TIM_SMCR_SMS_0*7;     // 7: Slave Trigger "Selected trigger clock the counter"
  TIM3->CCR1 = AdcCntPerSample/IRQPerSample; // allererster IRQ-Zeitpunkt
  TIM3->DIER = TIM_DIER_CC1IE;        // TIM3_IRQHandler
  TIM3->SR = ~TIM_SR_UIF;
  TIM3->CR1 = TIM_CR1_CEN;
  NVIC_EnableIRQ(TIM3_IRQn);

  // Init ADC
  my_assert(fMHz*1000*1000>32*fCarrier*ADCSamplesPerPeriod, _ErrAdcSampleTime); // ADC braucht mind. 32 CPU-Takte
  RCC->CFGR   |= RCC_CFGR_ADCPRE_0*0; // 0 -> PCLK2/2
  RCC->APB2ENR|= RCC_APB2ENR_ADC1EN;
  {
    // Calibrate ADC
    ADC1->CR2   |= ADC_CR2_ADON;        // ADC ON (0->1)
    ADC1->CR2   |= ADC_CR2_ADON;        // start (single) conversion (1->1)
    while((ADC1->SR&ADC_SR_EOC)==0);    // wait for the end of conversion
    ADC1->CR2   |= ADC_CR2_RSTCAL;      // reset calibration
    while((ADC1->CR2&ADC_CR2_RSTCAL));  // wait for reset
    ADC1->CR2   |= ADC_CR2_CAL;         // start calibration
    while((ADC1->CR2&ADC_CR2_CAL)!=0);  // wait till calibration is done
    ADC1->CR2   &= ~ADC_CR2_ADON;       // ADC OFF
  }
  ADC1->CR2   |= ADC_CR2_DMA;         // DMA Enabled
  ADC1->CR2   |= ADC_CR2_EXTTRIG;     // External trigger
  ADC1->CR2   |= ADC_CR2_EXTSEL_0*3;  // 3 -> Trigger TIM2_CC2
  ADC1->SQR1  |= ADC_SQR1_L_0*(1-1);  // 0 -> 1 Kanal
  ADC1->SQR3  |= ADC_SQR3_SQ1_0*7;    // 7 -> PA7
  ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*0;  // 0 -> 14 Takte (T0+Tx=12.5+1.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*1;  // 1 -> 20 Takte (T0+Tx=12.5+7.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*2;  // 2 -> 26 Takte (T0+Tx=12.5+13.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*3;  // 3 -> 41 Takte (T0+Tx=12.5+28.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*4;  // 4 -> 54 Takte (T0+Tx=12.5+41.5) fr PA7
  ADC1->CR2   |= ADC_CR2_ADON;        // ADC ON (0->1)
  ADC1->CR2   |= ADC_CR2_ADON;        // ADC Start (1->1) (->1ExtraSample, damit TIM3 synchron zum Sample)

  // Init DMA1_CH1 zum Lesen der ADC-Ergebnisse
  RCC->AHBENR |= RCC_AHBENR_DMA1EN;
  DMA1_Channel1->CNDTR = AdcBufCnt;
  DMA1_Channel1->CPAR  = (uint32_t)&ADC1->DR;
  DMA1_Channel1->CMAR  = (uint32_t)Control->TimedAdc.ValArr;
  DMA1_Channel1->CCR = (DMA_CCR1_MEM2MEM *0) |
                       (DMA_CCR1_PL_0    *0) |
                       (DMA_CCR1_MSIZE_0 *1) |
                       (DMA_CCR1_PSIZE_0 *1) |
                       (DMA_CCR1_MINC    *1) |
                       (DMA_CCR1_PINC    *0) |
                       (DMA_CCR1_CIRC    *1) |
                       (DMA_CCR1_DIR     *0) |
                       (DMA_CCR1_TEIE    *0) |
                       (DMA_CCR1_HTIE    *0) |
                       (DMA_CCR1_TCIE    *0) |
                       (DMA_CCR1_EN      *1);

  // Init DMA1_CH2 zum Setzen der ADC-Samplezeitpunktabstnde
  DMA1_Channel2->CNDTR = AdcBufCnt;
  DMA1_Channel2->CPAR  = (uint32_t)&TIM2->ARR;
  DMA1_Channel2->CMAR  = (uint32_t)Control->TimedAdc.TimArr; // Zeitabstnde
  DMA1_Channel2->CCR = (DMA_CCR1_MEM2MEM *0) |
                       (DMA_CCR1_PL_0    *0) |
                       (DMA_CCR1_MSIZE_0 *0) |
                       (DMA_CCR1_PSIZE_0 *1) |
                       (DMA_CCR1_MINC    *1) |
                       (DMA_CCR1_PINC    *0) |
                       (DMA_CCR1_CIRC    *1) |
                       (DMA_CCR1_DIR     *1) |
                       (DMA_CCR1_TEIE    *0) |
                       (DMA_CCR1_HTIE    *0) |
                       (DMA_CCR1_TCIE    *0) |
                       (DMA_CCR1_EN      *1);

  // alles in Gang setzen
  TIM2->CR1 = TIM_CR1_CEN;
}

//################################################### DDH47_Deinit
void OPTS DDH47_Deinit(TDDH47Control* s)
{
  // Stop
  TIM2->CR1 = 0;
  // Reset TIM2
  RCC->APB1RSTR |= RCC_APB1RSTR_TIM2RST;
  RCC->APB1RSTR &= ~RCC_APB1RSTR_TIM2RST;
  RCC->APB1ENR &= ~RCC_APB1ENR_TIM2EN;
  // Reset TIM3
  RCC->APB1RSTR |= RCC_APB1RSTR_TIM3RST;
  RCC->APB1RSTR &= ~RCC_APB1RSTR_TIM3RST;
  RCC->APB1ENR &= ~RCC_APB1ENR_TIM3EN;
  // Reset ADC1
  RCC->APB2RSTR |= RCC_APB2RSTR_ADC1RST;
  RCC->APB2RSTR &= ~RCC_APB2RSTR_ADC1RST;
  RCC->APB2ENR &= ~RCC_APB2ENR_ADC1EN;
  // Reset DMA1_Channel2
  DMA1_Channel2->CCR = 0;
  // Reset DMA1_Channel1
  DMA1_Channel1->CCR = 0;
  // Reset PA7
  GPIOA->CRL = (GPIOA->CRL&0x0fffffff)|0x40000000;
}
