
#include "stm32f10x.h"
#include "stdlib.h"
#include "stdarg.h"

#include "DDH47.h"

//
// ### Empfnger fr den Wettersender Pinneberg "DDH47" auf 147.3 MHz ###
//
// - Hardware
//   STM32F103C8 - "Blue Pill Development Board"
//
// - Anschlsse
//   GPIO-PA7: Antenneneingang: 147.3 MHz, mind. 0.1mV
//   GPIO-PA9: UART-Ausgang: "9600 8N1" (9600 Baud, ein Startbit, acht Datenbits, kein Paritybit, ein Stoppbit)
//
// - Abgleich
//   - Resonanzfrequenz von der Empfangsantenne gut abstimmen (+-1kHz)
//   - Antenne senkrecht zu Pinneberg ausrichten
//   - bei ausreichend Empfang ("SNR"-Wert ber 5) den ber UART angezeigten Wert fr "ppm"
//     fr "Crystal_PPM" (in "main.c") eintragen
//


const int  Crystal_PPM  = 0; // <<<< ### hier den ppm-Wert aus der UART-Ausgabe eintragen ###
const bool UART_DbgInfo = true;


//################################################### Set_HSE
static int Set_HSE(int MHz)
{
  // (Einweg - nicht wiederverwenden)

  // 1, 2, 4, 8
  //  8.. 64 in 4MHz Schritten
  // 64..136 in 8MHz Schritten

  // (schmutziger) Reset
  RCC->CFGR = 0;
  RCC->CR &= ~(RCC_CR_HSEON|RCC_CR_PLLON);
  FLASH->ACR &= ~FLASH_ACR_LATENCY;

  // HSE einschalten
  RCC->CR |= RCC_CR_HSEON;

  // warten bis HSE luft
  while((RCC->CR&RCC_CR_HSERDY)==0)
  {
  }

  // warten bis HSE Systemclock ist
  RCC->CFGR |= RCC_CFGR_SW_0*1;
  while((RCC->CFGR&RCC_CFGR_SWS)!=RCC_CFGR_SWS_0*1)
  {
  }

  if(MHz<=8)
  {
    if(MHz==- 16) RCC->CFGR |= RCC_CFGR_HPRE_DIV512; // kein STLINK Zugriff
    if(MHz==- 32) RCC->CFGR |= RCC_CFGR_HPRE_DIV256;
    if(MHz==- 64) RCC->CFGR |= RCC_CFGR_HPRE_DIV128;
    if(MHz==-125) RCC->CFGR |= RCC_CFGR_HPRE_DIV64;
    if(MHz==-500) RCC->CFGR |= RCC_CFGR_HPRE_DIV16;
    if(MHz==1) RCC->CFGR |= RCC_CFGR_HPRE_DIV8;
    if(MHz==2) RCC->CFGR |= RCC_CFGR_HPRE_DIV4;
    if(MHz==4) RCC->CFGR |= RCC_CFGR_HPRE_DIV2;
    if(MHz==8) RCC->CFGR |= RCC_CFGR_HPRE_DIV1;
    return 0;
  }

  // PLL auf HSE stellen
  RCC->CFGR |= RCC_CFGR_PLLSRC;

  // PLL braucht bei 8MHz ca. 0.1mA zu
  if(MHz>24)
  {
    // WaitStates
    if(MHz>48)
        FLASH->ACR |= FLASH_ACR_LATENCY_1*2;
      else
        FLASH->ACR |= FLASH_ACR_LATENCY_1*1;
  }
  if((MHz&4)!=0)
  {
    RCC->CFGR |= RCC_CFGR_PLLXTPRE;
    MHz *= 2;
  }
  RCC->CFGR = (RCC->CFGR&0xffc3ffff)|(((MHz>>3)-2)<<18);
  RCC->CR |= RCC_CR_PLLON;
  while((RCC->CR&RCC_CR_PLLRDY)==0)
  {
  }
  RCC->CFGR = (RCC->CFGR&~RCC_CFGR_SW)|(RCC_CFGR_SW_0*2);
  while((RCC->CFGR&RCC_CFGR_SWS)!=RCC_CFGR_SWS_0*2)
  {
  }
  return 0;
}

//################################################### uprintf
static void uprintf(const char* Fmt, ...)
{
  // "Mini-USART-printf()"
  void USART_out(char Char)
  {
    // TXD-Ausgang ist GPIO_PA9
    const int Baud = 9600;
    if((USART1->CR1&USART_CR1_UE)==0)
    {
      RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
      GPIOA->CRH = (GPIOA->CRH&0xffffff0f)|0x000000a0;
      RCC->APB2ENR |= RCC_APB2ENR_USART1EN;
      USART1->CR1 = USART_CR1_UE;
      int divval = (fMHz/1*1000000+Baud/2)/Baud; // 16..65535
      USART1->BRR = divval;
      USART1->CR1 |= USART_CR1_TE;
    }
    USART1->DR = Char;
    while((USART1->SR&USART_SR_TXE)==0)
    {
      __WFI();
    }
  }
  int GetNum(const char** Fmt)
  {
    int v = 0;
    while(**Fmt>='0' && **Fmt<='9')
    {
      v = v*10+(*(*Fmt)++-'0');
    }
    return v;
  }
  va_list va;
  va_start(va, Fmt);
  for(; *Fmt!=0; Fmt++)
  {
    if(*Fmt=='%')
      {
        bool Zero = *(++Fmt)=='0';
        int  n0 = GetNum(&Fmt);
        int  n1 = Zero ? n0 : 0;
        if(Zero) n1=n0;
        if(*Fmt=='.')
        {
          Fmt++;
          n1 = GetNum(&Fmt)+1;
        }
        switch(*Fmt)
        {
          case 'c': {
                      USART_out(va_arg(va, int32_t));
                      break;
                    }
          case 's': {
                      const char* s = va_arg(va, const char*);
                      if(n0!=0)
                      {
                        int n2 = 0;
                        for(; s[n2]!=0; n2++) { };
                        for(; n2<n0; n2++) USART_out(' ');
                      }
                      while(*s!=0) USART_out(*s++);
                      break;
                    }
          case 'i': {
                      int v = va_arg(va, int32_t);
                      int cv = 10;
                      int n2 = 1;
                      for(; abs(v)>=cv; n2++) cv*=10;
                      int nn = n1>n2 ? n1 : n2;
                      int nSpace = n0-nn;
                      if(v<0) nSpace--;
                      for(; nSpace>0; nSpace--) USART_out(' ');
                      if(v<0)
                      {
                        USART_out('-');
                        v = abs(v);
                      }
                      for(int i=0; i<nn; i++)
                      {
                        if(i>=n1-n2) cv/=10;
                        int n = v/cv;
                        USART_out(n+48);
                        v -= n*cv;
                        if(!Zero && i>=n2-n1)
                        {
                          USART_out('.');
                          Zero = true;
                        }

                      }
                      break;
                    }
        }
      }
    else
      USART_out(*Fmt);
  }
  va_end(va);
}

//################################################### main
int main(void)
{
  TDDH47State   DDH47State;
  TDDH47Control DDH47Control;

  Set_HSE(fMHz);
  DDH47_Init(&DDH47Control, &DDH47State, Crystal_PPM);

  // GPIO_PC13 (LED) auf Ausgang (50MHz)
  RCC->APB2ENR |= RCC_APB2ENR_IOPCEN;
  GPIOC->CRH = (GPIOC->CRH&0xff0fffff)|0x00300000;
  GPIOC->BSRR = 1<<(13+0*16); // LED aus

  for(int iLoop=0; ; iLoop++)
  {
    __WFI();

    // empfangenes Zeichen
    // volatile char* Code = &DDH47State.TriteDecode; // einfacher Dekoder
    volatile char* Code = &DDH47State.FancyDecode; // alternativ Spezialdekoder

    // lesen und ausgeben
    if(*Code!=0)
    {
      uprintf("%c", *Code);

      // Debug-Ausgaben
      if(UART_DbgInfo && *Code==10)
      {
        uprintf("\r%8i %8i uV  %4i SNR  %3i ppm %s> ",
                DDH47State.HalfBitCnt,
                DDH47State.SN_uV,
                DDH47State.SNR,
                DDH47State.dfHSE_ppm,
                abs(Crystal_PPM-DDH47State.dfHSE_ppm)>50 ? " - Crystal_PPM anpassen! " : "");
      }
      *Code = 0;
    }
  }
  return 0;
}
